/*
 * File:  def.c
 * Date:  September 14, 1993
 * Programmer:  Cecil Huang
 * =========================
 * This file exports the basic function interfaces and type definitions
 * used in the source code for SWIG1.12.
 */


#include <stdio.h>


/* General I/O stuff */

#define PROGRAM_NAME         "SWIG1.12"
#define MAXLINE              10240
#define MAXSTRING            256
#define NUMVARS              16
#define MAXFILENAME          256
#define TRUE                 -1
#define FALSE                0
#define INFINITY             1.0E+12


/*
 * Mappings from parameter strings to integers.  Code is sensitive
 * to these number assignments, so don't change!
 */

#define ALGO               0
#define PEP1_FILE          1
#define PEP2_FILE          2
#define GAP_PENALTY_INTERP 3
#define PAM_FILE           4
#define DGAP_VARY          5
#define DGAP_FILE          6
#define QGAP_VARY          7
#define QGAP_FILE          8
#define OUTFILE            9
#define MATFILE           10


/* These must be non-zero. */

#define SMITH_WATERMAN     1
#define NEEDLEMAN_WUNSCH   2
#define PROFILE            3

#define CONSTANT           1
#define VARIABLE           2

#define ABSOLUTE           1
#define NORM_WEIGHTED      2
#define NORM_UNWEIGHTED    3


/* PAM Data Type */

#define NOT -1

typedef struct pam {
  int size;               /* number of amino acid types                */
  int **table;            /* array of pam values                       */
  char *aa_list;          /* array of amino acid characters            */
  
  int *hash;              /* hash array for getting amino acid indices */
  int hash_subtract;      /* (int) AA_letter - AA_hash_table_index     */
  int hash_size;          /* size of hash array                        */
} PAM;


/* PROTEIN Data Type */

#define MAXNAME 512

typedef struct protein {
  char name[MAXNAME]; /* name of protein                               */
  int length;         /* number of amino acids                         */
  char *sequence;     /* amino acid sequence.  1st aa at position ONE. */
} PROTEIN;

PROTEIN *make_protein();
void free_protein(PROTEIN *protein);


/*
 * CELL Data Type
 * --------------
 * Note that, effective in MAT03, the first peptide is still indexed by i,
 * and the second peptide is still indexed by j.  However, there are now
 * i COLUMNS and j ROWS, in correspondence with the literature out there.
 */

#define OFFSET 3          /* new for sw1.05 -- say something about this */
#define PTR_MASK 7        /* mask in the last 3 bits */
#define JDEC  4
#define IDEC  2
#define IJDEC 1
#define NONE  0

typedef struct cell {
  double value;           /* Value at this cell                        */
  int    ptr;             /* Indicates pointer to derivative cell      */

  double p_value;         /* Value of gap penalty in i direction       */
  int    p_length;        /* Gap size in i direction                   */

  double q_value;         /* Value of gap penalty in j direction       */
  int    q_length;        /* Gap size in j direction                   */
} CELL;

CELL **make_CELL_table(int rows, int columns);
void free_CELL_table(CELL **mat, int rows);


/*
 * MAT_PLUS Data Type
 * ------------------
 * The MAT_PLUS Data Type packages a table with its auxiliary information.
 */

typedef struct mat_plus {
  CELL            **mat;       /* Best match pathway matrix            */
  int             **mat_align; /* just the backtrace information       */
  int             mat_align_rows;
  double          max_value;   /* value of the best match              */
  int             rmax, cmax;  /* position of the best match           */
  PROTEIN         *p_database, *p_query;    /* Proteins used           */
} MAT_PLUS;
  

MAT_PLUS *make_MAT_PLUS(struct protein *p_database,
			struct protein *p_query);
void free_MAT_PLUS(MAT_PLUS *mp);


/* SSTRING (stack_string) Data Type */

typedef struct stack_string {
  char *string;           /* stack of characters                       */
  int pos;                /* stack pointer                             */
} SSTRING;

SSTRING *make_sstring(int len);
void free_sstring(SSTRING *k);
void ss_append(char c, SSTRING *ss);
char *ss_string(SSTRING *ss);


/* Integers */

int *make_int_array(int len);
int *extend_int_array(int *array, int newlen);


/* Characters */

char *make_char_array(int len);
char *extend_char_array(char *array, int newlen);
int getline (FILE *fp, char *line, int max);
int gap_penalty_interp_index (char *s);
