/*
 * File:  io.h 
 * Date:  September 14, 1993
 * Programmer:  Cecil Huang
 * =========================
 * I/O implementation for SWIG1.12
 * 
 * September 16, 1993.  Revamped read_protein.  Forget dynamic allocation.
 */


#include <ctype.h>
#include "def.h"
#include "io.h"
#include <stdio.h>
#include <stdlib.h>


static char scratch[MAXLINE];


/* for use by print_optimal_match () */

extern char outfile[MAXLINE];


/* for use by read_and_calculate_gap_penalties */

extern char
  qgap_vary[MAXLINE], qgap_file[MAXLINE],
  dgap_vary[MAXLINE], dgap_file[MAXLINE],
  gap_penalty_interp[MAXLINE];

extern double
  qgap_penalty, qgap_size_penalty,
  dgap_penalty, dgap_size_penalty;


/* Function Prototypes */

static int pam_aa_index ();
static int get_aa (FILE *fp);


int **make_int_table (int rows, int columns)
{
  int **table, i;

  table = (int **) malloc (rows * sizeof (int *));

  for (i = 0; i < rows; i++) {
    table[i] = (int *) malloc (columns * sizeof (int));
  }

  return (table);
}


void free_int_table (int **mat_align, int rows)
{
  int i;

  for (i = 0; i < rows; i++) {
    free (mat_align[i]);
  }

  free (mat_align);
}


/*
 * ---> PAM I/O Routines (NCBI matrices) <---
 *
 * read_pam (filename)
 *     returns a pointer to a PAM which contains information read in
 *     from <filename>.
 *
 * make_pam ()
 *     returns a pointer to a new, empty PAM structure.
 *
 * pam_lookup (pam, aa1, aa2)
 *     returns the log-odds similarity of <aa1> and <aa2>, as recorded
 *     in the PAM pointed to by <pam>.
 *
 * print_pam (pam)
 *     prints the PAM pointed to by <pam>, in a format similar to that
 *     of the files created by pam.c (see below).
 *
 * pam_copy(dest, src)
 *     makes a copy of the PAM pointed to by <src>.  <dest> points to
 *     the new PAM.
 */

PAM *read_pam (char filename[])
{
  FILE *fp;
  PAM *pam;
  char *aa_list, c, max, min;
  int i, j, x, size, **table, *hash;
  
  
  fp = fopen (filename, "r"); /* error checking already occurred */

  while (((i = getline (fp, scratch, MAXLINE)) > 0) 
	 && (scratch[0] == '#')) {
  }

  if (i == 0) {
    fprintf (stdout, "Error in matrix file: %s\n", filename);
    fclose (stdout);

    exit (-1);
  }


  /* count number of characters in first line */

  x = strlen (scratch);

  for (i = size = 0; i < x; i++)
    if (isgraph ((int) scratch[i])) {
      /* printing character except space */
      size++;
    }

  pam = make_pam ();
  pam->size = size;
  pam->table = table = make_int_table (size, size);
  pam->aa_list = aa_list = make_char_array (size);

  for (i = j = 0; i < x; i++) {
    if (isgraph ((int) c = scratch[i])) {
      c = tolower (c);
      aa_list[j] = c;
      j++;
    }
  }

  for (min = max = aa_list[0], i = 1; i < size; i++) {
    x = aa_list[i];

    if (x > max) {
      max = x;
    } else if (x < min) {
      min = x;
    }
  }

  pam->hash_subtract = (int) min;
  pam->hash_size = x = (int) (max - min + 1);
  pam->hash = hash = make_int_array (x);

  for (i = 0; i < x; i++) {
    hash[i] = NOT;
  }

  for (i = 0; i < size; i++) {
    hash[(int) (aa_list[i] - min)] = i;
  }

  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      if (fscanf(fp, "%d", &table[i][j]) != 1) {
	fprintf (stdout, "Error in matrix file: %s\n", filename);
	fclose (stdout);

	exit (-1);
      }
    }
  }
    
  fclose (fp);
  return pam;
}

void free_pam (PAM *pam)
{
  int i, size;

  size = pam->size;

  for (i = 0; i < size; i++) {
    free (pam->table[i]);
  }

  free (pam->table);
  free (pam->hash);
  free (pam->aa_list);

  free (pam);
}


void print_pam (PAM *pam)
{
  int size, i, j, **table;


  size = pam->size;
  table = pam->table;

  for (i = 0; i < size; i++) {
    printf ("  %c",toupper (pam->aa_list[i]));
  }

  printf ("\n");

  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      printf ("%3d", table[i][j]);
    }

    printf ("\n");
  }
}
      

PAM *make_pam ()
{
  return ((PAM *) malloc (sizeof (PAM)));
}


int pam_lookup (PAM *pam, char aa1, char aa2)
{
  return (pam->table[pam_aa_index (pam, aa1)][pam_aa_index (pam, aa2)]);
}


static int pam_aa_index (PAM *pam, int aa)
{
  int i;

  i = (int) (aa - pam->hash_subtract);
  return ((i >= 0) && (i < pam->hash_size)) ? pam->hash[i] : NOT;
}


void pam_copy (PAM *dest, PAM *src)
{
  int i, j, size, hash_size;
  
  dest->size = size = src->size;
  dest->hash_subtract = src->hash_subtract;
  dest->hash_size = hash_size = src->hash_size;

  dest->table = make_int_table (size, size);

  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      dest->table[i][j] = src->table[i][j];
    }
  }

  dest->aa_list = make_char_array (size);

  for (i = 0; i < size; i++) {
    dest->aa_list[i] = src->aa_list[i];
  }

  dest->hash = make_int_array(hash_size);

  for (i = 0; i < hash_size; i++) {
    dest->hash[i] = src->hash[i];
  }
}


/*
 * ---> PROTEIN I/O Routines <---
 *
 * read_protein (pam, filename)
 *     returns a pointer to a PROTEIN, which contains an amino acid
 *     sequence read from the peptide file <filename>.  <pam> is used
 *     as an error-checking facility:  if any amino acids are not
 *     present in the PAM pointed to by <pam>, an error is flagged.
 *
 * print_protein (protein)
 *     prints the peptide pointed to by <protein> in ig format.  An
 *     example of this format can be found in the file "test.pep",
 *     currently located in /a/huang/src/mat01/examples.
 *
 * protein->name          returns name of protein   (char *)
 * protein->length        returns length of protein (int)
 * protein->sequence      returns protein sequence  (char *)
 */
    
PROTEIN *read_protein (PAM *pam, char filename[])
{
  FILE    *fp;
  int     i, aa_count;
  int     c;
  char *cptr;
  PROTEIN *protein;
  

  fp = fopen (filename, "r");
  i = 0;

  while ((c = fgetc(fp)) != EOF) {
    i++;
  }

  fclose (fp);

  fp = fopen (filename, "r");
  while (((getline (fp, scratch, MAXLINE)) > 0) 
	 && (scratch[0] == ';')) {
  }

  protein                  = make_protein ();
  protein->sequence = cptr = make_char_array (i);
  sscanf (scratch, "%s", protein->name);
  aa_count = 0;
  *cptr++ = '0';

  while (pam_aa_index (pam, (c = get_aa (fp))) != NOT) {
    aa_count++;
    *cptr++ = c;
  }

  switch (c) {
    case EOF:
      fprintf (stderr, "%s: %s: premature end of file.\n",
	       PROGRAM_NAME, filename);
      fclose (stderr);
      abort ();
    case ';':
      fprintf (stderr, "%s: %s: premature end of protein.\n",
	       PROGRAM_NAME, filename);
      fclose (stderr);
      abort ();
    case '1':
    case '2':
      protein->length = aa_count;
      fclose (fp);
      return (protein);
    default:
      fprintf (stderr, "%s: %s: amino acid %c in position %d not valid.\n",
	       PROGRAM_NAME, filename, c, (aa_count + 1));
      fprintf (stderr, "pam_aa_index is %d\n", pam_aa_index (pam, c));
      fclose (stderr);

      abort ();
    }
}


void print_protein (PROTEIN *protein)
{
  int len, i;
  char *cptr;


  len = protein -> length;

  for (i = 0, cptr = protein->sequence + 1; i < len; i++, cptr++) {
    if (i && !(i % 70)) {
      printf ("\n");
    }

    printf ("%c", *cptr);
  }

  printf ("\n");
}


/*
 * Gap Penalty Calculation Routines
 *
 *     read_and_calculate_gap_penalties(pam, p1, p2)
 *         opens query gap and database gap files and obtains
 *         user-specified values.  Then converts to values for internal
 *         use, depending on PAM values, amino acid composition, and
 *         user-specified gap-penalty-interpretation mode (ABSOLUTE,
 *         NORMALIZED-WEIGHTED, NORMALIZED-UNWEIGHTED).
 */

void read_and_calculate_gap_penalties (PAM *pam, PROTEIN *p1, PROTEIN *p2)
{
  double
    user_qgap_penalty,
    user_qgap_size_penalty,
    user_dgap_penalty,
    user_dgap_size_penalty,
    gap_conv_factor, sum, top, bottom;
  FILE *fp;
  int i, gpi, num_aa, fxf, **pam_table, *p1_freqs, *p2_freqs;


  fp = fopen (qgap_file, "r");
  if (!getline (fp, scratch, MAXLINE) ||
      (sscanf (scratch, "%lf %lf",
	       &user_qgap_penalty, &user_qgap_size_penalty) != 2)) {
    fprintf (stderr, "Error reading query gap penalty file \"%s\"\n",
	     qgap_file);
    fclose (stderr);

    abort ();
  }

  fclose (fp);
  
  fp = fopen (dgap_file, "r");

  if (!getline (fp, scratch, MAXLINE) ||
      (sscanf (scratch, "%lf %lf",
	      &user_dgap_penalty, &user_dgap_size_penalty) != 2)) {
    fprintf (stderr, "Error reading database gap penalty file \"%s\"\n",
	     dgap_file);
    fclose (stderr);

    abort ();
  }

  fclose (fp);

  gpi = gap_penalty_interp_index (gap_penalty_interp);

  if (gpi == ABSOLUTE) {
    qgap_penalty      = user_qgap_penalty;
    qgap_size_penalty = user_qgap_size_penalty;
    dgap_penalty      = user_dgap_penalty;
    dgap_size_penalty = user_dgap_size_penalty;
    return;
  }

  num_aa = pam->size;
  pam_table = pam->table;

  if (gpi == NORM_UNWEIGHTED) {
    for (i = 0, sum = 0.0; i < num_aa; i++)
      sum += pam_table[i][i];

    gap_conv_factor   = sum / num_aa;
    qgap_penalty      = gap_conv_factor * user_qgap_penalty;
    qgap_size_penalty = qgap_penalty * user_qgap_size_penalty;
    dgap_penalty      = gap_conv_factor * user_dgap_penalty;
    dgap_size_penalty = dgap_penalty * user_dgap_size_penalty;
    return;
  }

  /* gpi == NORM_WEIGHTED */

  p1_freqs = get_freq_data (pam, p1);
  p2_freqs = get_freq_data (pam, p2);

  top = bottom = 0.0;

  for (i = 0; i < num_aa; i++) {
    fxf = p1_freqs[i] * p2_freqs[i];
    top = top + fxf * pam_table[i][i];
    bottom = bottom + fxf;
  }

  free (p1_freqs);
  free (p2_freqs);

  gap_conv_factor  = top / bottom;

  qgap_penalty      = gap_conv_factor * user_qgap_penalty;
  qgap_size_penalty = qgap_penalty * user_qgap_size_penalty;
  dgap_penalty      = gap_conv_factor * user_dgap_penalty;
  dgap_size_penalty = dgap_penalty * user_dgap_size_penalty;
  return;
}


int *get_freq_data (PAM *pam, PROTEIN *protein)
{
  int i, pep_len, num_aa, *freq_array;
  char *pep_seq;


  num_aa = pam->size;
  freq_array = make_int_array(num_aa);

  for (i = 0; i < num_aa; i++) {
    freq_array[i] = 0;
  }

  pep_seq = protein->sequence;
  pep_len = protein->length;

  for (i = 1; i <= pep_len; i++) {
    freq_array[pam_aa_index(pam, pep_seq[i])]++;
  }

  return freq_array;
}


/*
 *
 * ---> MAT_PLUS I/O Routines <---
 *
 * print_optimal_match(mp)
 *     Generates and prints an optimal alignment by tracing back through
 *     the direction matrix.  "In retracing a path from the E matrix, it
 *     is necessary to continue with the previous direction if possible,
 *     else to prefer a diagonal or vertical segment to a horizontal
 *     one." (Taylor, p. 450.)  By following this protocol, a path with
 *     as few turns as possible is traced.
 *
 *     Relevant BLAZE-type statistics are also generated.
 */

void x_beginning(char *top, char *bot)
{
  int i, top_Xable, bot_Xable;


  top_Xable = (top[1] == ' ');
  bot_Xable = (bot[1] == ' ');

  if (top_Xable && bot_Xable) {
    top[0] = 'X';
    bot[0] = 'X';
  } else if (top_Xable) {
    top[0] = 'X';
  } else if (bot_Xable) {
    bot[0] = 'X';
  } else {
    top[0] = 'X';
    
    for (i = 1; isdigit ((int) top[i]); i++) {
      top[i] = ' ';
    }
  }
}


void x_end(char *top, char *bot, int len)
{
  int i, top_Xable, bot_Xable;

  top_Xable = (top[len - 2] == ' ');
  bot_Xable = (bot[len - 2] == ' ');

  if (top_Xable && bot_Xable) {
    top[len - 1] = 'X';
    bot[len - 1] = 'X';
  } else if (top_Xable) {
    top[len - 1] = 'X';
  } else if (bot_Xable) {
    bot[len - 1] = 'X';
  } else {
    top[len - 1] = 'X';

    for (i = len - 2; isdigit ((int) top[i]); i--) {
      top[i] = ' ';
    }
  }
}


void print_optimal_match (MAT_PLUS *mp, PAM *pam)
{
  int 
    **mat_align, current, num_match, num_mismatch, num_cons_sub, num_gaps,
    rmax, cmax, maxlen, i, j, k, dec, ptr, prev_ptr, align_len, pd_len;
  SSTRING *s1, *s2, *align, *top, *bottom;
  double  score, best_score;
  char    
    *database_seq, *query_seq, c1, c2,
    *str_top, *str_bottom, *str_align, *str_s1, *str_s2;
  FILE    *fp;


  mat_align    = mp->mat_align;
  maxlen       = mp->p_database->length + mp->p_query->length;
  rmax         = mp->rmax;
  cmax         = mp->cmax;
  database_seq = mp->p_database->sequence;
  query_seq    = mp->p_query->sequence;
  s1           = make_sstring(maxlen);
  s2           = make_sstring(maxlen);
  align        = make_sstring(maxlen);
  top          = make_sstring(maxlen);
  bottom       = make_sstring(maxlen);

  num_match = num_mismatch = num_cons_sub = num_gaps = 0;


/*
 * The percent match is relative to the best_score.  The best_score
 * is the score obtained when the database sequence is aligned with itself.
 */

  pd_len = mp->p_database->length;
  best_score = 0.0;

  for (i = 1; i <= pd_len; i++) {
    c1 = database_seq[i];
    best_score += pam_lookup (pam, c1, c1);
  }
  

/*
 * Here, we perform the backtracing necessary to generate the alignment.
 */

  for (i = rmax, j = cmax, prev_ptr = IJDEC; i && j; ) {

    current = mat_align[i][j];
    ptr = current & PTR_MASK;

    if (ptr & prev_ptr) {
      ptr = prev_ptr;
    } else {
      ptr = prev_ptr = (ptr & IJDEC) ? IJDEC :
	((ptr & IDEC) ? IDEC : ((ptr & JDEC) ? JDEC : NONE));
    }

    switch (ptr) {
      
    case IDEC:
      dec = current >> OFFSET;
      for (k = 0; k < dec; k++) {
	ss_append ('-', s1);
	ss_append (toupper (query_seq[i-k]), s2);

	ss_append (next_bottom_char (i - k), bottom);
	ss_append (next_top_char (1), top);

	ss_append (' ', align);
      }

      i  -= dec;
      num_gaps += dec;
      break;   
    case JDEC:
      dec = current >> OFFSET;

      for (k = 0; k < dec; k++) {
	ss_append ('-', s2);
	ss_append (toupper (database_seq[j-k]), s1);

	ss_append (next_bottom_char(1), bottom);
	ss_append (next_top_char(j - k), top);

	ss_append (' ', align);
      }

      j  -= dec;
      num_gaps += dec;
      break;
    case IJDEC:
      c2 = query_seq[i];
      ss_append (toupper (c2), s2);
      c1 = database_seq[j];
      ss_append (toupper (c1), s1);

      if (c2 == c1) {
	ss_append('|', align);
	num_match++;
      } else if (pam_lookup (pam, c2, c1) > 0) {
	ss_append (':', align);
	num_cons_sub++;
      } else {
	ss_append (' ', align);
	num_mismatch++;
      }

      ss_append (next_bottom_char(i),    bottom);
      ss_append (next_top_char(j), top);

      i--;
      j--;
      break;
    case NONE:
      i = 0;
      break;
    }
  }


/*
 * What follows is a glop of code which prints out the results
 * in a pleasant format, similar to that of BLAZE.
 */

  fp = (outfile[0] == '\0') ? stdout : fopen(outfile, "w");

  fprintf (fp, "%s makes an optimal alignment between two sequences.\n\n",
	   PROGRAM_NAME);
  fprintf (fp, "(  TOP )  Database peptide: %12s (length %5d)\n",
	   mp->p_database->name, mp->p_database->length);
  fprintf (fp, "(BOTTOM)  Query    peptide: %12s (length %5d)\n",
	   mp->p_query->name, mp->p_query->length);

  fprintf (fp, "\n");
  
  score = mp->max_value;                /* new code -- long overdue */
  fprintf (fp, "Score = %5.2f  (Perfect = %5d)                      ",
	   score, (int) (best_score + 0.5));
  fprintf (fp, "Percent Match = %3d%%\n",
	   (int) ((score * 100.0/best_score) + 0.5));

  fprintf (fp, "Residue Identity =   %3d%%  ",
	   (int) (num_match * 100.0 /
		(num_match + num_mismatch + num_cons_sub + num_gaps)));
  fprintf (fp, "Matches          = %6d  ", num_match);
  fprintf (fp, "Mismatches   =%6d\n", num_mismatch);
  fprintf (fp, "Gaps             = %6d  ", num_gaps);
  fprintf 
    (fp, "Conservative Substitutions              =%6d\n\n", num_cons_sub);

  str_bottom    = ss_string (bottom);
  str_top = ss_string (top);
  str_s2     = ss_string (s2);
  str_s1     = ss_string (s1);
  str_align  = ss_string (align);

  align_len = strlen (str_align); /* length the same for all strings */

  x_beginning (str_bottom, str_top           );
  x_end       (str_bottom, str_top, align_len);

  do {
    fprintf (fp, "    %.72s\n", str_top);
    fprintf (fp, "    %.72s\n", str_s1    );
    fprintf (fp, "    %.72s\n", str_align );
    fprintf (fp, "    %.72s\n", str_s2    );
    fprintf (fp, "    %.72s\n", str_bottom   );
    fprintf (fp,"\n");

    str_top += 72;
    str_s1 += 72;
    str_align += 72;
    str_s2 += 72;
    str_bottom += 72;
    align_len -= 72;
  } while (align_len > 0);

  free_sstring(s1);
  free_sstring(s2);
  free_sstring(align);
  free_sstring(top);
  free_sstring(bottom);

  fclose(fp);
}


char next_top_char(int aa_pos)
{
  static char stack[16];
  static int  stack_pos = -1;
  
  if (stack_pos >= 0) {
    return (stack[stack_pos--]);
  }

  if ((aa_pos % 10) == 0) {
    sprintf (stack, "%d", aa_pos);
    stack_pos = strlen (stack) - 1;
    return (stack[stack_pos--]);
  }

  return (' ');
}


char next_bottom_char (int aa_pos)
{
  static char stack[16];
  static int  stack_pos = -1;
  
  if (stack_pos >= 0) {
    return (stack[stack_pos--]);
  }

  if ((aa_pos % 10) == 0) {
    sprintf (stack, "%d", aa_pos);
    stack_pos = strlen (stack) - 1;
    return (stack[stack_pos--]);
  }

  return (' ');
}


/* kind of like a buffered getchar, but CRs are skipped */

static int get_aa (FILE *fp) 
{
  int c;

  while (!isgraph (c = getc (fp)) && (c != EOF)) {
  }

  return (tolower (c)); /* this MUST be lower case for pam_hash */
}
