/*
 * File:  main.c
 * Date:  September 14, 1993
 * Programmer:  Cecil Huang
 * =========================
 * This is the top-level routine for SWIG1.12.  SWIG stands for Smith-
 * Waterman Intelligenetics.  This implementation is functionally identical
 * to SWIG1.11, but with the following changes:
 *
 *  1.  The matrix is filled row by row instead of column by column.
 *
 *  2.  As a result, the write_edit_matrix feature is restored.
 */


#include "def.h"
#include "io.h"
#include "params.h"
#include <stdio.h>
#include <stdlib.h>
#include "sw.h"


/*
 * These are the parameters that the user can input.  The module
 * params.c reads and checks these parameters.
 */

char
  algo[MAXLINE],
  database_pepfile[MAXLINE],
  query_pepfile[MAXLINE],
  gap_penalty_interp[MAXLINE],
  matrix_file[MAXLINE],

  dgap_vary[MAXLINE],
  dgap_file[MAXLINE],

  qgap_vary[MAXLINE],
  qgap_file[MAXLINE],

  outfile[MAXLINE],
  matfile[MAXLINE];


/*
 * gap penalties -- needs revamping for PROFILE algorithm
 */

double
  qgap_penalty, qgap_size_penalty,
  dgap_penalty, dgap_size_penalty;


/* Function Prototypes */

static void help(char *s);


int main (int argc, char **argv)
{
  MAT_PLUS *mp;
  PROTEIN *p_database, *p_query;
  PAM *pam;
  
  if (argc != 2) {
    help(argv[0]);
  }

  parse_and_check_params (argv[1]);

  pam = read_pam (matrix_file);
  p_database = read_protein (pam, database_pepfile);
  p_query    = read_protein (pam, query_pepfile);
  read_and_calculate_gap_penalties (pam, p_database, p_query);

  /*
   * changes in gap penalty interpretation, as described above,
   * are hacked in below.
   */

  qgap_penalty = qgap_penalty - qgap_size_penalty;    /* new code */
  dgap_penalty = dgap_penalty - dgap_size_penalty;    /* new code */

  mp = make_MAT_PLUS (p_database, p_query);

  switch (algo_index (algo)) {
  case SMITH_WATERMAN:
    smith_waterman_affine_max (mp, pam);
    break;
  case NEEDLEMAN_WUNSCH:
    needleman_wunsch_affine_max (mp, pam);
    break;
  default:
    fprintf (stderr, "Internal error.\n");
    fclose (stderr);
    abort ();
  }

  print_optimal_match (mp, pam);

  free_MAT_PLUS (mp);
  free_pam (pam);
  free_protein (p_database);
  free_protein (p_query);

  exit (0);
}


static void help(char *s)
{
  fprintf (stdout, "-------------------------------------------------------------------------------\n");
  fprintf (stdout, "%s Usage:  %s <parameter-file>\n", PROGRAM_NAME, s);
  fprintf (stdout, "-------------------------------------------------------------------------------\n");
  fprintf (stdout, "Parameter file format is as follows:\n\n");
  fprintf (stdout, "algorithm             = {SMITH-WATERMAN, NEEDLEMAN-WUNSCH, PROFILE}\n");
  fprintf (stdout, "database_file         = <peptide1-file-path>\n");
  fprintf (stdout, "query_file            = <peptide2-file-path>\n");
  fprintf (stdout, "gap_penalty           = {ABSOLUTE, NORMALIZED-WEIGHTED, NORMALIZED-UNWEIGHTED}\n");
  fprintf (stdout, "matrix_file              = <pam-file-path>\n");
  fprintf (stdout, "database_gap_option   = {CONSTANT, VARIABLE}\n");
  fprintf (stdout, "database_gap_file     = <database-gap-file-path>\n");
  fprintf (stdout, "query_gap_option      = {CONSTANT, VARIABLE}\n");
  fprintf (stdout, "query_gap_file        = <query-gap-file-path>\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "# Blank lines and lines beginning with '#' are ignored.\n");
  fprintf (stdout, "# The following lines are optional.\n");
  fprintf (stdout, "\n");
  fprintf (stdout, "output_alignment_file = <output-file-path>\n");
  fprintf (stdout, "edit_matrix_file      = <edit-matrix-file-path>\n");
  fprintf (stdout, "\n");

  fclose (stdout);
  exit (0);
}
