/*
 * File:  sw.c
 * Date:  September 14, 1993
 * Programmer:  Cecil Huang
 * =========================
 * This module contains routines related to the dynamic table used by
 * SWIG1.12 to determine the best match pathway.
 *
 * BUG FIX:  3/8/1993
 * I used -1.0 as a dummy initializer.  But I had p_value and
 * q_value switched around, so that at the edges, the algorithm was
 * actually comparing it to -1.0!!  (Fixed for nw_initialize_table too.)
 */


#include "def.h"
#include "io.h"
#include <stdio.h>
#include <stdlib.h>
#include "sw.h"


#define DUMMY -9.9


/*
 * Gap penalties are a function of the PAM matrix, the sequence
 * composition, and the user-specified gap-penalty-interpretation flag.
 * They are calculated in io.c.
 */
  
extern double
  qgap_penalty, qgap_size_penalty,
  dgap_penalty, dgap_size_penalty;

extern char matfile[MAXLINE];


static double calculate_p (CELL **mat, int i, int j);
static double calculate_q (CELL **mat, int i, int j);
static void sw_initialize_table (MAT_PLUS *mp);


static void sw_initialize_table (MAT_PLUS *mp)
{
  int      i, j, rows, columns;
  CELL     **mat, *cell;
  int      **mat_align;


  mat       = mp->mat;
  mat_align = mp->mat_align;   /** new code **/
  rows      = mp->p_query->length;
  columns   = mp->p_database->length;

  mat[0][0].value = 0.0;
  mat[0][0].ptr   = NONE;

  mat_align[0][0] = NONE;      /** new code **/

  for (i = 1; i <= rows; i++) {
    mat_align[i][0]   = NONE;
  }

  for (j = 1; j <= columns; j++) {
    mat_align[0][j]   = NONE;
  }

  /* initialize left hand side of row 1 */

  cell = &mat[1][0];
  cell->value = cell->q_value = 0.0;
  cell->p_value     = DUMMY;
  cell->ptr         = NONE;
  
  /* initialize 0th row in every column */

  for (j = 1; j <= columns; j++) {
    cell = &mat[0][j];
    cell->value = cell->p_value = 0.0;
    cell->q_value     = DUMMY;
    cell->ptr         = NONE;
  }
}  


static double calculate_p (CELL **mat, int i, int j)
{
  CELL   *this_cell, *prev_cell;
  double d_sim, p_sim;


  this_cell = &mat[i][j];
  prev_cell = &mat[i-1][j];

  d_sim = prev_cell->value - qgap_penalty;
  p_sim = prev_cell->p_value;

  if (d_sim > p_sim) {
    this_cell->p_length = 1;
    return (this_cell->p_value = d_sim - qgap_size_penalty);
  } else {
    this_cell->p_length = prev_cell->p_length + 1;
    return (this_cell->p_value = p_sim - qgap_size_penalty);
  }
}


static double calculate_q (CELL **mat, int i, int j)
{
  CELL   *this_cell, *prev_cell;
  double d_sim, q_sim;
  

  this_cell = &mat[i][j];
  prev_cell = &mat[i][j-1];

  d_sim = prev_cell->value - dgap_penalty;
  q_sim = prev_cell->q_value;

  if (d_sim > q_sim) {
    this_cell->q_length = 1;
    return (this_cell->q_value = d_sim - dgap_size_penalty);
  } else {
    this_cell->q_length = prev_cell->q_length + 1;
    return (this_cell->q_value = q_sim - dgap_size_penalty);
  }
}


void smith_waterman_affine_max (MAT_PLUS *mp, PAM *pam)
{
  int 
    i, j, ptr, rows, columns, rmax, cmax,
    dec, write_matfile, **mat_align;
  char    *database_seq, *query_seq;
  CELL    **mat, *cell, *temp;
  double  max, p_value, q_value, ijdec_value, table_max;
  FILE    *matfp;

  sw_initialize_table (mp);

  mat          = mp->mat;
  mat_align    = mp->mat_align;
  database_seq = mp->p_database->sequence;    
  query_seq    = mp->p_query->sequence;

  rows         = mp->p_query->length;
  columns      = mp->p_database->length;

  table_max    = 0.0;
  rmax         = 0;
  cmax         = 0;

  write_matfile = (matfile[0] == '\0') ? FALSE : TRUE;
  matfp = NULL;

  if (write_matfile) {
    matfp = fopen (matfile, "w");
    fwrite (&rows   , sizeof (int), 1, matfp);
    fwrite (&columns, sizeof (int), 1, matfp);
  }

  for (i = 1; i <= rows; i++) {
    for (j = 1; j <= columns; j++) {
        p_value     = calculate_p (mat, 1, j);
	q_value     = calculate_q (mat, 1, j);

	ijdec_value = mat[0][j-1].value
	  + pam_lookup(pam, query_seq[i], database_seq[j]);

        /* Now see which is the best.  Note the bitwise OR.  */
        ptr     = NONE;
	max     = 0.0;

	if (ijdec_value > max) { 
	  /* strictly > since we want SUBsequences */
	  ptr = IJDEC;
	  max = ijdec_value;
	}

	if (p_value >= max) {
	  ptr = (p_value == max) ? ptr | IDEC : IDEC;
	  max = p_value;
	}

	if (q_value >= max) {
	  ptr = (q_value == max) ? ptr | JDEC : JDEC;
	  max = q_value;
	}

        cell = &mat[1][j];
	cell->value   = max;
	cell->ptr     = ptr;

	if (max > table_max) {
	  table_max = max;
	  rmax = i;
	  cmax = j;
	}
      }

    if (write_matfile) {
      for (j = 1; j <= columns; j++) {
	fwrite (&(mat[1][j].value), sizeof (double), 1, matfp);
      }
    }
      
    /* update mat_align array after each row */

    for (j = 1; j <= columns; j++) {
      cell = &mat[1][j];
      ptr  = cell->ptr;

      if (ptr == IDEC) {
	dec = cell->p_length;
      } else if (ptr == JDEC) {
	dec = cell->q_length;
      } else {
	dec = 0;
      }

      mat_align[i][j] = (dec << OFFSET) + ptr;
    }

    /* exchange row 0 and row 1, reinitialize row 1 */

    temp = mat[0];
    mat[0] = mat[1];
    mat[1] = temp;

    cell = &mat[1][0];
    cell->value = cell->q_value = 0.0;
    cell->p_value = DUMMY;
    cell->ptr = NONE;
  }

  mp->max_value = table_max;
  mp->rmax      = rmax;
  mp->cmax      = cmax;

  if (write_matfile && matfp != NULL) {
    fclose (matfp);
  }
}


void nw_initialize_table (MAT_PLUS *mp)
{
  int      i, j, rows, columns, **mat_align;
  CELL     **mat, *cell;
  double   penalty;


  mat       = mp->mat;
  mat_align = mp->mat_align;
  rows      = mp->p_query->length;
  columns   = mp->p_database->length;

  mat[0][0].value = 0.0;
  mat[0][0].ptr   = NONE;

  /* new code */

  mat_align[0][0] = NONE;

  for (i = 1; i <= rows; i++) {
    mat_align[i][0] = NONE;
  }

  for (j = 1; j <= columns; j++) {
    mat_align[0][j] = NONE;
  }
  
  /* initialize left hand side of row 1 */

  cell = &mat[1][0];
  cell->value = cell->q_value = -(qgap_penalty + qgap_size_penalty);
  cell->p_value     = DUMMY;
  cell->ptr         = NONE;
  
  /* initialize 0th row in every column */

  penalty = dgap_penalty + dgap_size_penalty;

  for (j = 1; j <= columns; j++, penalty += dgap_size_penalty) {
    cell = &mat[0][j];
    cell->value = cell->p_value = -penalty;
    cell->q_value     = DUMMY;
    cell->ptr         = NONE;
  }
}


void needleman_wunsch_affine_max (MAT_PLUS *mp, PAM *pam)
{
  int
    i, j, ptr, rows, columns, rmax, 
    cmax, dec, write_matfile, **mat_align;
  char    *database_seq, *query_seq;
  CELL    **mat, *cell, *temp;
  double  max, p_value, q_value, ijdec_value;
  FILE    *matfp;


  nw_initialize_table (mp);

  mat          = mp->mat;
  mat_align    = mp->mat_align;
  database_seq = mp->p_database->sequence;    
  query_seq    = mp->p_query->sequence;

  rows         = mp->p_query->length;
  columns      = mp->p_database->length;
  rmax         = 0;
  cmax         = 0;

  write_matfile = (matfile[0] == '\0') ? FALSE : TRUE;
  matfp = NULL;

  if (write_matfile) {
    matfp = fopen (matfile, "w");
    fwrite (&rows   , sizeof (int), 1, matfp);
    fwrite (&columns, sizeof (int), 1, matfp);
  }

  for (i = 1; i <= rows; i++) {
    for (j = 1; j <= columns; j++) {
        p_value     = calculate_p (mat, 1, j);
	q_value     = calculate_q (mat, 1, j);

	ijdec_value = mat[0][j-1].value +
	  pam_lookup (pam, query_seq[i], database_seq[j]);

        /* 
	 * Now see which is the best.  Note the bitwise OR.
	 * NOTE HERE that we do not require 0 to be a mininum.
         * As per Needleman-Wunsch, this allows negative numbers to be
	 * in the matrix.
	 */

	ptr = IJDEC;
	max = ijdec_value;
	
	if (p_value >= max) {
	  ptr = (p_value == max) ? ptr | IDEC : IDEC;
	  max = p_value;
	}

	if (q_value >= max) {
	  ptr = (q_value == max) ? ptr | JDEC : JDEC;
	  max = q_value;
	}

        cell = &mat[1][j];
	cell->value   = max;
	cell->ptr     = ptr;
      }

    if (write_matfile) {
      for (j = 1; j <= columns; j++) {
	fwrite (&(mat[1][j].value), sizeof(double), 1, matfp);
      }
    }

    /* update mat_align array after each row */

    for (j = 1; j <= columns; j++) {
      cell = &mat[1][j];
      ptr  = cell->ptr;

      if (ptr == IDEC) {
	dec = cell->p_length;
      } else if (ptr == JDEC) {
	dec = cell->q_length;
      } else {
	dec = 0;
      }

      mat_align[i][j] = (dec << OFFSET) + ptr;
    }

    /* exchange row 0 and row 1, reinitialize row 1 */

    temp = mat[0];
    mat[0] = mat[1];
    mat[1] = temp;

    /* Here, we are initializing row i+1 effectively */

    cell = &mat[1][0];
    cell->value =
      cell->q_value = -(qgap_penalty + (i+1) * qgap_size_penalty);
    cell->p_value = DUMMY;
    cell->ptr     = NONE;
  }

  /* 
   * we want max_value to be the bottom-right corner.
   * this would be row 0, last column (remember rows got swapped)
   */
     
  mp->max_value = mat[0][columns].value;
  mp->rmax      = rows;
  mp->cmax      = columns;
}
